from xml.etree import ElementTree

import pytest

# Apply module-level marks
pytestmark = pytest.mark.integration


###############################################################################
# TEST DATA ###################################################################
###############################################################################
PRELIMINARY_VOEVENT_DATA = {
    'skymap_type': None,
    'skymap_filename': None,
    'internal': True,
    'hardware_inj': False,
    'open_alert': False,
    'CoincComment': False,
    'ProbHasNS': 0.1,
    'ProbHasRemnant': 0.9,
    'BNS': 0.2,
    'NSBH': 0.3,
    'BBH': 0.4,
    'Terrestrial': 0.5,
    'MassGap': 0.6,
}

INITIAL_VOEVENT_DATA = {
    'skymap_type': 'new',
    'skymap_filename': 'fake_skymap.txt',
    'internal': True,
    'hardware_inj': False,
    'open_alert': False,
    'CoincComment': False,
    'ProbHasNS': 0.1,
    'ProbHasRemnant': 0.9,
    'BNS': 0.2,
    'NSBH': 0.3,
    'BBH': 0.4,
    'Terrestrial': 0.5,
    'MassGap': 0.6,
}


###############################################################################
# UTILITY FUNCTIONS ###########################################################
###############################################################################
def get_citations_dict(obj_id, voevent_file_text):
    # Parse XML
    xml_data = ElementTree.fromstring(voevent_file_text)
    citations_dict = {}
    for ei in xml_data.findall('./Citations/EventIVORN'):
        citations_dict[ei.text] = ei.attrib['cite']
    return citations_dict


###############################################################################
# TESTS #######################################################################
###############################################################################
@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_preliminary_voevent(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create VOEvent and check results
    response = client.createVOEvent(obj_id, 'PR', **PRELIMINARY_VOEVENT_DATA)
    assert response.status_code == 201
    data = response.json()
    assert data['voevent_type'] == 'PR'
    assert data['skymap_type'] == PRELIMINARY_VOEVENT_DATA['skymap_type']
    assert data['skymap_filename'] == \
        PRELIMINARY_VOEVENT_DATA['skymap_filename']
    assert data['internal'] == PRELIMINARY_VOEVENT_DATA['internal']
    assert data['hardware_inj'] == PRELIMINARY_VOEVENT_DATA['hardware_inj']
    assert data['open_alert'] == PRELIMINARY_VOEVENT_DATA['open_alert']
    assert data['coinc_comment'] == PRELIMINARY_VOEVENT_DATA['CoincComment']
    assert data['prob_has_ns'] == PRELIMINARY_VOEVENT_DATA['ProbHasNS']
    assert data['prob_has_remnant'] == \
        PRELIMINARY_VOEVENT_DATA['ProbHasRemnant']
    assert data['prob_bns'] == PRELIMINARY_VOEVENT_DATA['BNS']
    assert data['prob_nsbh'] == PRELIMINARY_VOEVENT_DATA['NSBH']
    assert data['prob_bbh'] == PRELIMINARY_VOEVENT_DATA['BBH']
    assert data['prob_terrestrial'] == PRELIMINARY_VOEVENT_DATA['Terrestrial']
    assert data['prob_mass_gap'] == PRELIMINARY_VOEVENT_DATA['MassGap']


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_initial_voevent(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Upload a fake skymap file
    response = client.writeLog(
        obj_id, "Fake skymap file",
        filename=INITIAL_VOEVENT_DATA['skymap_filename'],
        filecontents="Fake skymap."
    )
    assert response.status_code == 201

    # Create a preliminary VOEvent to update
    response = client.createVOEvent(obj_id, 'PR', **PRELIMINARY_VOEVENT_DATA)
    assert response.status_code == 201
    preliminary_data = response.json()

    # Create VOEvent and check results
    response = client.createVOEvent(obj_id, 'IN', **INITIAL_VOEVENT_DATA)
    assert response.status_code == 201
    data = response.json()
    assert data['voevent_type'] == 'IN'
    assert data['skymap_type'] == INITIAL_VOEVENT_DATA['skymap_type']
    assert data['skymap_filename'] == INITIAL_VOEVENT_DATA['skymap_filename']
    assert data['internal'] == INITIAL_VOEVENT_DATA['internal']
    assert data['hardware_inj'] == INITIAL_VOEVENT_DATA['hardware_inj']
    assert data['open_alert'] == INITIAL_VOEVENT_DATA['open_alert']
    assert data['coinc_comment'] == INITIAL_VOEVENT_DATA['CoincComment']
    assert data['prob_has_ns'] == INITIAL_VOEVENT_DATA['ProbHasNS']
    assert data['prob_has_remnant'] == INITIAL_VOEVENT_DATA['ProbHasRemnant']
    assert data['prob_bns'] == INITIAL_VOEVENT_DATA['BNS']
    assert data['prob_nsbh'] == INITIAL_VOEVENT_DATA['NSBH']
    assert data['prob_bbh'] == INITIAL_VOEVENT_DATA['BBH']
    assert data['prob_terrestrial'] == INITIAL_VOEVENT_DATA['Terrestrial']
    assert data['prob_mass_gap'] == INITIAL_VOEVENT_DATA['MassGap']

    # Check citations
    response = client.files(obj_id, data['filename'])
    assert response.status_code == 200
    voevent_file_text = response.read()
    citations_dict = get_citations_dict(obj_id, voevent_file_text)
    assert len(citations_dict) == 1
    assert preliminary_data['ivorn'] in citations_dict
    assert citations_dict[preliminary_data['ivorn']] == 'supersedes'


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_update_voevent(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Upload a fake skymap file
    response = client.writeLog(
        obj_id, "Fake skymap file",
        filename=INITIAL_VOEVENT_DATA['skymap_filename'],
        filecontents="Fake skymap."
    )
    assert response.status_code == 201

    # Create a preliminary VOEvent to update
    response = client.createVOEvent(obj_id, 'PR', **PRELIMINARY_VOEVENT_DATA)
    assert response.status_code == 201
    preliminary_data = response.json()

    # Create an initial VOEvent to update
    response = client.createVOEvent(obj_id, 'IN', **INITIAL_VOEVENT_DATA)
    assert response.status_code == 201
    initial_data = response.json()

    # Create VOEvent and check results
    response = client.createVOEvent(obj_id, 'UP', **INITIAL_VOEVENT_DATA)
    assert response.status_code == 201
    data = response.json()
    assert data['voevent_type'] == 'UP'
    assert data['skymap_type'] == INITIAL_VOEVENT_DATA['skymap_type']
    assert data['skymap_filename'] == INITIAL_VOEVENT_DATA['skymap_filename']
    assert data['internal'] == INITIAL_VOEVENT_DATA['internal']
    assert data['hardware_inj'] == INITIAL_VOEVENT_DATA['hardware_inj']
    assert data['open_alert'] == INITIAL_VOEVENT_DATA['open_alert']
    assert data['coinc_comment'] == INITIAL_VOEVENT_DATA['CoincComment']
    assert data['prob_has_ns'] == INITIAL_VOEVENT_DATA['ProbHasNS']
    assert data['prob_has_remnant'] == INITIAL_VOEVENT_DATA['ProbHasRemnant']
    assert data['prob_bns'] == INITIAL_VOEVENT_DATA['BNS']
    assert data['prob_nsbh'] == INITIAL_VOEVENT_DATA['NSBH']
    assert data['prob_bbh'] == INITIAL_VOEVENT_DATA['BBH']
    assert data['prob_terrestrial'] == INITIAL_VOEVENT_DATA['Terrestrial']
    assert data['prob_mass_gap'] == INITIAL_VOEVENT_DATA['MassGap']

    # Check citations
    response = client.files(obj_id, data['filename'])
    assert response.status_code == 200
    voevent_file_text = response.read()
    citations_dict = get_citations_dict(obj_id, voevent_file_text)
    assert len(citations_dict) == 2
    assert preliminary_data['ivorn'] in citations_dict
    assert initial_data['ivorn'] in citations_dict
    assert citations_dict[preliminary_data['ivorn']] == 'supersedes'
    assert citations_dict[initial_data['ivorn']] == 'supersedes'


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_retraction_voevent(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Upload a fake skymap file
    response = client.writeLog(
        obj_id, "Fake skymap file",
        filename=INITIAL_VOEVENT_DATA['skymap_filename'],
        filecontents="Fake skymap."
    )
    assert response.status_code == 201

    # Create a preliminary VOEvent to retract
    response = client.createVOEvent(obj_id, 'PR', **PRELIMINARY_VOEVENT_DATA)
    assert response.status_code == 201
    preliminary_data = response.json()

    # Create a retraction VOEvent
    response = client.createVOEvent(obj_id, 'RE')
    assert response.status_code == 201
    retraction_data = response.json()
    assert retraction_data['voevent_type'] == 'RE'

    # Check citations
    response = client.files(obj_id, retraction_data['filename'])
    assert response.status_code == 200
    voevent_file_text = response.read()
    citations_dict = get_citations_dict(obj_id, voevent_file_text)
    assert len(citations_dict) == 1
    assert preliminary_data['ivorn'] in citations_dict
    assert citations_dict[preliminary_data['ivorn']] == 'retraction'



import os
import pytest
from ligo.gracedb.exceptions import HTTPError

# Apply module-level marks
pytestmark = pytest.mark.integration


def test_update_grbevent(client, test_data_dir):
    """Test update GRB event"""
    # Setup: create a GRB event
    event_file = os.path.join(test_data_dir, 'fermi-test.xml')
    response = client.createEvent('Test', 'Fermi', event_file, search='GRB')
    assert response.status_code == 201
    initial_data = response.json()
    gid = initial_data['graceid']

    # Update the grbevent's parameters
    redshift = 3.4
    designation = 'very good'
    ra = 12.34
    try:
        response = client.update_grbevent(
            gid,
            redshift=redshift,
            designation=designation,
            ra=ra
        )
        new_data = response.json()

        # Even though they're test GRB events, unprivileged users
        # can't update GRB events. So try to catch a 200 error
        # in the case of an admin or grb user, and 403 otherwise.
        # This came up with setting up the gitlab integration instance.

        assert response.status_code == 200
        # Compare results
        initial_grb_params = initial_data['extra_attributes']['GRB']
        new_grb_params = new_data['extra_attributes']['GRB']
        assert new_grb_params['ra'] == ra
        assert new_grb_params['redshift'] == redshift
        assert new_grb_params['designation'] == designation
        assert new_grb_params['ra'] != initial_grb_params['ra']
        assert new_grb_params['redshift'] != initial_grb_params['redshift']

    except HTTPError as e:
        assert e.status_code == 403



# Tests for subcommands below 'get'
#  Ex: 'gracedb get superevent'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_get_emobservation_subcommand(CLI):
    """Test get emobservation subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'N': 14,
    }

    # Generate command
    cmd = 'get emobservation {s_id} {N}'.format(**cmd_args)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.emobservations'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'],)
    assert cli_kwargs == {'emobservation_num': cmd_args['N']}


def test_get_event_subcommand(CLI):
    """Test get event subcommand"""
    g_id = 'G1234'

    # Generate command
    cmd = 'get event {g_id}'.format(g_id=g_id)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.event'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (g_id,)
    assert cli_kwargs == {}


FILE_TEST_DATA = [None, 'test.txt', '/file/we/cannot/access.txt']
@pytest.mark.parametrize("dest", FILE_TEST_DATA)  # noqa: E302
def test_get_file_subcommand(CLI, dest):
    """Test get file subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'filename': 'fake_filename.txt',
    }

    # Generate command
    cmd = 'get file {s_id} {filename}'.format(**cmd_args)
    if dest is not None:
        cmd += ' {dest}'.format(dest=dest)

    func = 'ligo.gracedb.cli.client.CommandLineClient.files'
    with mock.patch(func) as mock_cli_func:
        if dest and 'cannot' in dest:
            # Last case should raise an IOError
            with pytest.raises(IOError):
                CLI(shlex.split(cmd))
            return
        else:
            CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['filename'],)
    assert cli_kwargs == {}


def test_get_label_subcommand(CLI):
    """Test get label subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'label_name': 'FAKE_LABEL',
    }

    # Generate command
    cmd = 'get label {s_id} {label_name}'.format(**cmd_args)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.labels'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'],)
    assert cli_kwargs == {'label': cmd_args['label_name']}


def test_get_log_subcommand(CLI):
    """Test get log subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'N': 12,
    }

    # Generate command
    cmd = 'get log {s_id} {N}'.format(**cmd_args)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.logs'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'],)
    assert cli_kwargs == {'log_number': cmd_args['N']}


SIGNOFF_TEST_DATA = [None, 'FAKE_INSTRUMENT']
@pytest.mark.parametrize("instrument", SIGNOFF_TEST_DATA)  # noqa: E302
def test_get_signoff_subcommand(CLI, instrument):
    """Test get signoff subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'signoff_type': 'FAKE_SIGNOFF_TYPE',
    }

    # Generate command
    cmd = 'get signoff {s_id} {signoff_type}'.format(**cmd_args)
    if instrument is not None:
        cmd += " {inst}".format(inst=instrument)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.signoffs'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'],)
    assert cli_kwargs == {'signoff_type': cmd_args['signoff_type'],
                          'instrument': instrument or ''}


def test_get_superevent_subcommand(CLI):
    """Test get superevent subcommand"""
    s_id = 'S001122a'

    # Generate command
    cmd = 'get superevent {s_id}'.format(s_id=s_id)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.superevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id,)
    assert cli_kwargs == {}


def test_get_voevent_subcommand(CLI):
    """Test get voevent subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'N': 12,
    }

    # Generate command
    cmd = 'get voevent {s_id} {N}'.format(**cmd_args)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.voevents'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'],)
    assert cli_kwargs == {'voevent_num': cmd_args['N']}



import datetime
try:
    from unittest import mock
except ImportError:
    import mock

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
import pytest

from ligo.gracedb.rest import GraceDb


@pytest.fixture
def safe_client():
    """A client class which has its request() method mocked away"""
    client = GraceDb()
    client.request = mock.Mock()
    return client


@pytest.fixture
def x509_key():
    """X.509 private key"""
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    return key


@pytest.fixture
def x509_cert(x509_key):
    """Self-signed X.509 certificate, valid for 1 hour"""
    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"Wisconsin"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, u"Milwaukee"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
        x509.NameAttribute(NameOID.COMMON_NAME, u"mysite.com"),
    ])
    cert = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        x509_key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(
        # Our certificate will be valid for 1 hour
        datetime.datetime.utcnow() + datetime.timedelta(seconds=3600)
    ).add_extension(
        x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
        critical=False,  # noqa: E122
    # Sign our certificate with our private key
    ).sign(x509_key, hashes.SHA256(), default_backend())

    return cert



import pytest

# Apply module-level marks
pytestmark = pytest.mark.integration


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_logs(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create a log
    comment = 'test comment'
    response = client.writeLog(obj_id, comment)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    log_N = data['N']

    # Pull down list of logs for the event
    response = client.logs(obj_id)
    assert response.status_code == 200
    data = response.json()
    assert len(data['log']) >= 1

    # Pull down individual log
    response = client.logs(obj_id, log_N)
    assert response.status_code == 200
    data = response.json()
    assert data['comment'] == comment
    assert data['N'] == log_N


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_log_creation_with_tags(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create a log with tags
    comment = 'test log with tags'
    tags = ['test_tag1', 'test_tag2']
    response = client.writeLog(obj_id, comment, tag_name=tags)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    for t in tags:
        assert t in data['tag_names']


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_log_tag_and_untag(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create a log
    comment = 'test log, add/remove tags later'
    response = client.writeLog(obj_id, comment)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    log_N = data['N']

    # Add a tag
    tag = 'test_tag'
    response = client.addTag(obj_id, log_N, tag)
    assert response.status_code == 201

    # Get log and check tag status
    response = client.logs(obj_id, log_N)
    assert response.status_code == 200
    data = response.json()
    assert len(data['tag_names']) == 1
    assert tag in data['tag_names']

    # Remove tag
    response = client.removeTag(obj_id, log_N, tag)
    assert response.status_code == 204

    # Get log and check tag status
    response = client.logs(obj_id, log_N)
    assert response.status_code == 200
    data = response.json()
    assert len(data['tag_names']) == 0



# Custom exceptions
import requests.exceptions


class HTTPError(requests.exceptions.HTTPError):
    def __init__(self, response):
        self.response = response
        self.status_code = response.status_code
        self.status = response.status_code
        self.reason = response.reason
        self.message = response.text
        self.text = response.text
        Exception.__init__(self, response)



# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Tanner Prestegard,
#               Branson Stephens, Brian Moe (2020)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>

# Sources:
#  1) https://stackoverflow.com/questions/45539422/can-we-reload-a-page-url-
#     in-python-using-urllib-or-urllib2-or-requests-or-mechan

#  2) https://2.python-requests.org/en/master/user/advanced/#example-
#     specific-ssl-version

#  3) https://urllib3.readthedocs.io/en/1.2.1/pools.html


from functools import partial
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPSConnection
from urllib3.connectionpool import HTTPSConnectionPool, HTTPConnectionPool
from .cert import check_certificate_expiration


class GraceDbCertAdapter(HTTPAdapter):
    def __init__(self, cert=None, reload_buffer=0, **kwargs):
        super(GraceDbCertAdapter, self).__init__(**kwargs)
        https_pool_cls = partial(
            GraceDbCertHTTPSConnectionPool,
            cert=cert,
            reload_buffer=reload_buffer)

        self.poolmanager.pool_classes_by_scheme = {
            'http': HTTPConnectionPool,
            'https': https_pool_cls
        }


class GraceDbCertHTTPSConnection(HTTPSConnection):
    def __init__(self, host, cert=None, reload_buffer=0, **kwargs):
        # At this point, te HTTPSConnection is initialized
        # but unconnected. Set this property to 'True'
        self.unestablished_connection = True
        super(GraceDbCertHTTPSConnection, self).__init__(host, **kwargs)

    @property
    def unestablished_connection(self):
        return self._unestablished_connection

    @unestablished_connection.setter
    def unestablished_connection(self, value):
        self._unestablished_connection = value

    def connect(self):
        # Connected. After this step, the unestablished
        # property is false.
        self.unestablished_connection = False
        super(GraceDbCertHTTPSConnection, self).connect()


class GraceDbCertHTTPSConnectionPool(HTTPSConnectionPool):
    # ConnectionPool object gets used in the HTTPAdapter.
    # "ConnectionCls" is a HTTP(S)COnnection object to use
    # As the underlying connection.

    # Source: https://urllib3.readthedocs.io/en/latest/
    #         reference/#module-urllib3.connectionpool

    ConnectionCls = GraceDbCertHTTPSConnection

    def __init__(self, host, port=None, cert=None,
                 reload_buffer=0, **kwargs):

        super(GraceDbCertHTTPSConnectionPool, self).__init__(
            host, port=port, **kwargs)

        self._cert = cert
        self._reload_buffer = reload_buffer

    def _expired_cert(self):
        return check_certificate_expiration(
            self._cert,
            self._reload_buffer)

    def _get_conn(self, timeout=None):
        while True:
            # Start the connection object. At this step, the connection
            # unestablished variable is true
            conn = super(GraceDbCertHTTPSConnectionPool, self)._get_conn(
                timeout)

            # 'returning' the connection object then triggers the
            # connection to be established. Establish a new connection
            # if it's unestablished, or if the cert expiration is within the
            # reload buffer. Establishing the new connection will (hopefully
            # load the new cert.
            if conn.unestablished_connection or not self._expired_cert():
                return conn

            # otherwise, kill the connection which will reset unestablished_..
            # to true and then exit the loop.
            conn.close()



import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import object_id_parser, superevent_id_parser


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class AddCommand(RegisteredCommandBase):
    name = "add"
    description = textwrap.dedent("""\
        Add a label to an event or superevent, a tag to a log
        entry, or an event to a superevent
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class AddChildBase(RegisteredSubCommandBase):
    _registry = registry


class AddEventCommand(AddChildBase):
    name = "event"
    description = "Add an event to a superevent"
    long_description = textwrap.dedent("""\
        Add an event to a superevent. They must fall in the same 'category';
        i.e., production, test, or MDC. Special permissions are required for
        non-Test events and superevents.
    """).rstrip()
    parent_parsers = (superevent_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            'graceid',
            type=str,
            help='Graceid of event to add to superevent'
        )
        return parser

    def run(self, client, args):
        return client.addEventToSuperevent(args.superevent_id, args.graceid)


class AddLabelCommand(AddChildBase):
    name = "label"
    description = "Add a label to an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "label",
            type=str,
            help=("Name of label to apply (do '{prog} info labels' to see "
                  "options)").format(prog=self.base_prog)
        )
        return parser

    def run(self, client, args):
        return client.writeLabel(args.object_id, args.label)


class AddTagCommand(AddChildBase):
    name = "tag"
    description = "Add a tag to an event or superevent log entry"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "log_number",
            metavar="N",
            type=int,
            help="Index number of log entry"
        )
        parser.add_argument(
            "tag_name",
            type=str,
            help="Name of tag to apply")
        parser.add_argument(
            "-d",
            "--tag-display-name",
            type=str,
            help="Display name for tag (new tags only)"
        )
        return parser

    def run(self, client, args):
        return client.addTag(
            args.object_id, args.log_number, tag_name=args.tag_name,
            displayName=args.tag_display_name
        )



import datetime

import pytest

# Apply module-level marks
pytestmark = pytest.mark.integration


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_emobservations(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Compile EM observation data
    emgroup = client.em_groups[0]
    ra_list = [1, 2, 3, 4]
    ra_width_list = [0.5] * len(ra_list)
    dec_list = [5, 6, 7, 8]
    dec_width_list = [0.7] * len(dec_list)
    now = datetime.datetime.utcnow()
    start_time_list = list(
        map(lambda i: (now + datetime.timedelta(seconds=i)).isoformat(),
            [0, 1, 2, 3])
    )
    duration_list = [1] * len(start_time_list)
    comment = "test comment"

    # Create EM observation and check results
    response = client.writeEMObservation(
        obj_id, emgroup, ra_list, ra_width_list, dec_list, dec_width_list,
        start_time_list, duration_list, comment=comment
    )
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    assert data['group'] == emgroup
    assert len(data['footprints']) == len(ra_list)
    for emf in data['footprints']:
        N = emf['N']
        assert emf['ra'] == ra_list[N - 1]
        assert emf['dec'] == dec_list[N - 1]
        assert emf['raWidth'] == ra_width_list[N - 1]
        assert emf['decWidth'] == dec_width_list[N - 1]
        assert emf['exposure_time'] == duration_list[N - 1]
    emo_N = data['N']

    # Get list of emobservations and check results
    response = client.emobservations(obj_id)
    assert response.status_code == 200
    data = response.json()
    assert len(data['observations']) == 1
    assert data['observations'][0]['N'] == emo_N

    # Retrieve the individual emobservation directly
    response = client.emobservations(obj_id, emo_N)
    assert response.status_code == 200
    data = response.json()
    assert data['comment'] == comment
    assert data['group'] == emgroup
    assert len(data['footprints']) == len(ra_list)
    for emf in data['footprints']:
        N = emf['N']
        assert emf['ra'] == ra_list[N - 1]
        assert emf['dec'] == dec_list[N - 1]
        assert emf['raWidth'] == ra_width_list[N - 1]
        assert emf['decWidth'] == dec_width_list[N - 1]
        assert emf['exposure_time'] == duration_list[N - 1]



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize("emo_number", [None, 12])
def test_superevent_emobservations(safe_client, emo_number):
    superevent_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if emo_number:
        template_key = 'superevent-emobservation-detail-template'
    else:
        template_key = 'superevent-emobservation-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.emobservations(superevent_id,
                                   emobservation_num=emo_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if emo_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['superevent_id'] == superevent_id
    if emo_number:
        assert template_call_kwargs['N'] == emo_number


@pytest.mark.parametrize("emo_number", [None, 12])
def test_event_logs(safe_client, emo_number):
    graceid = 'T123456'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if emo_number:
        template_key = 'emobservation-detail-template'
    else:
        template_key = 'emobservation-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.emobservations(graceid, emobservation_num=emo_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if emo_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['graceid'] == graceid
    if emo_number:
        assert template_call_kwargs['N'] == emo_number


@pytest.mark.parametrize(
    "is_event,obj_id",
    [
        (True, "T123456"),
        (False, "TS121212abc"),
    ]
)
def test_write_emobservations(safe_client, is_event, obj_id):

    # Generate data
    emgroup = 'FAKE_EMGROUP'
    ra_list = [1] * 4
    ra_width_list = [2] * 4
    dec_list = [3] * 4
    dec_width_list = [4] * 4
    start_time_list = [5] * 4
    duration_list = [6] * 4
    comment = 'test'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'emobservation-list-template'
    else:
        template_key = 'superevent-emobservation-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'em-groups': [emgroup]}

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.writeEMObservation(
            obj_id, emgroup, ra_list, ra_width_list, dec_list, dec_width_list,
            start_time_list, duration_list, comment=comment
        )

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 1
    assert 'json' in post_call_kwargs
    request_body = post_call_kwargs['json']
    assert request_body['group'] == emgroup
    assert request_body['ra_list'] == ra_list
    assert request_body['ra_width_list'] == ra_width_list
    assert request_body['dec_list'] == dec_list
    assert request_body['dec_width_list'] == dec_width_list
    assert request_body['start_time_list'] == start_time_list
    assert request_body['duration_list'] == duration_list
    assert request_body['comment'] == comment

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    if is_event:
        call_key = 'graceid'
    else:
        call_key = 'superevent_id'
    assert template_call_kwargs[call_key] == obj_id


@pytest.mark.parametrize(
    "ras,ra_widths,decs,dec_widths,start_times,durations",
    [
        ([1] * 4, [1] * 4, [1] * 4, [1] * 4, [1] * 4, [1] * 5),
        ([1] * 4, [1] * 4, [1] * 4, [1] * 4, [1] * 5, [1] * 4),
        ([1] * 4, [1] * 4, [1] * 4, [1] * 5, [1] * 4, [1] * 4),
        ([1] * 4, [1] * 4, [1] * 5, [1] * 4, [1] * 4, [1] * 4),
        ([1] * 4, [1] * 5, [1] * 4, [1] * 4, [1] * 4, [1] * 4),
        ([1] * 5, [1] * 4, [1] * 4, [1] * 4, [1] * 4, [1] * 4),
        ([1] * 1, [1] * 2, [1] * 3, [1] * 4, [1] * 5, [1] * 6),
    ]
)
def test_write_emobservation_different_list_lengths(
    safe_client, ras, ra_widths, decs, dec_widths, start_times, durations
):

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'em-groups': ['emgroup1']}

    err_msg = ("raList, decList, startTimeList, raWidthList, decWidthList, "
               "and durationList should be the same length")
    with mock.patch(si_prop, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client.writeEMObservation(
                'TS121212a', 'emgroup1', ras, ra_widths, decs, dec_widths,
                start_times, durations, comment="test"
            )


def test_write_emobservation_bad_emgroup(safe_client):

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'em-groups': ['emgroup1']}

    err_msg = "group must be one of {groups}".format(
        groups=", ".join(mock_si_dict['em-groups']))
    with mock.patch(si_prop, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client.writeEMObservation(
                'TS121212a', 'emgroup20', 1, 1, 1, 1, 1, 1, comment="test"
            )



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


def test_get_superevent_permissions(safe_client):
    superevent_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-permission-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.permissions(superevent_id)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    assert template_call_kwargs['superevent_id'] == superevent_id


def test_get_event_permissions(safe_client):
    graceid = 'T123456'

    err_msg = 'Not implemented for events'
    with pytest.raises(NotImplementedError, match=err_msg):
        safe_client.permissions(graceid)


@pytest.mark.parametrize("action", ['expose', 'hide'])
def test_modify_superevent_permissions(safe_client, action):
    superevent_id = 'TS121212a'

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = \
        {'superevent-permission-modify-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.modify_permissions(superevent_id, action)

    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 1
    assert 'data' in post_call_kwargs
    assert post_call_kwargs['data'] == {'action': action}

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    assert template_call_kwargs['superevent_id'] == superevent_id


@pytest.mark.parametrize("action", ['expose', 'hide'])
def test_modify_event_permissions(safe_client, action):
    graceid = 'T123456'
    err_msg = 'Not implemented for events'
    with pytest.raises(NotImplementedError, match=err_msg):
        safe_client.modify_permissions(graceid, action)


@pytest.mark.parametrize("action", ['other', 1, 3.4, False, (), ['hide']])
def test_modify_permissions_bad_action(safe_client, action):
    superevent_id = 'TS121212abc'
    err_msg = "action should be 'expose' or 'hide'"
    with pytest.raises(ValueError, match=err_msg):
        safe_client.modify_permissions(superevent_id, action)



import re
import six
import mimetypes
import requests
from functools import wraps
from .exceptions import HTTPError

EVENT_PREFIXES = ['G', 'E', 'H', 'M', 'T']
event_prefix_regex = re.compile(r'^({prefixes})\d+'.format(
    prefixes="|".join(EVENT_PREFIXES)))


# Decorator for class methods so that they work for events or superevents
def event_or_superevent(func):
    @wraps(func)
    def inner(self, object_id, *args, **kwargs):
        is_superevent = True
        if event_prefix_regex.match(object_id):
            is_superevent = False
        return func(self, object_id, is_superevent=is_superevent, *args,
                    **kwargs)
    return inner


# Function for checking arguments which can be strings or lists
# If a string, converts to list for ease of processing
def handle_str_or_list_arg(arg, arg_name):
    if arg:
        if isinstance(arg, six.string_types):
            arg = [arg]
        elif isinstance(arg, list):
            pass
        else:
            raise TypeError("{0} arg is {1}, should be str or list"
                            .format(arg_name, type(arg)))
    return arg


# Convert a dictionary into a list of tuples. This is used when uploading
# MultipartEncoded form data (basically uploading with files). The encoding
# fails when converting dicts of lists {'fruit': ['apple', 'orange']} into
# form-encoded data.
# Source: https://toolbelt.readthedocs.io/en/latest/
#     uploading-data.html#requests_toolbelt.multipart.encoder.MultipartEncoder
def dict_to_form_encoded(body):
    fields = []
    for k, v in body.items():
        if isinstance(v, list):
            for item in v:
                fields.append((k, item))
        else:
            fields.append((k, v))
    return fields


# Guess and return the mimetype of a given file. Otherwise return
# application/octect-stream
def get_mimetype(file_to_send):
    mtype, mencoding = mimetypes.guess_type(file_to_send)
    if not mtype:
        mtype = 'application/octet-stream'
    return mtype


# Add a response "hook" that adds old compatibility methods to request's
# httpresonse. source:
# https://requests.readthedocs.io/en/master/user/advanced/#event-hooks
def hook_response(r, *args, **kwargs):
    r.status = r.status_code


# Another hook: return the raise_for_status behavior which was previously
# enabled in the adjustResponse function.
def raise_status_exception(r, *args, **kwargs):
    try:
        r.raise_for_status()
    except requests.HTTPError as e:
        raise HTTPError(response=e.response)



import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import object_id_parser, superevent_id_parser

# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class ListCommand(RegisteredCommandBase):
    name = "list"
    description = textwrap.dedent("""\
        List files, log entries, labels, EM observations, signoffs, or VOEvents
        associated with an event or superevent, or list tags attached to a log
        entry
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class ListChildBase(RegisteredSubCommandBase):
    _registry = registry
    parent_parsers = (object_id_parser,)
    client_func = None

    def run(self, client, args):
        if self.client_func is not None:
            func = getattr(client, self.client_func)
        else:
            func = getattr(client, self.name)

        # Get either object_id, superevent_id, or graceid from args
        object_id_arg = None
        for arg_id in ['object_id', 'superevent_id', 'graceid']:
            if hasattr(args, arg_id):
                object_id_arg = getattr(args, arg_id)
                break
        return func(object_id_arg)


class ListEmobservationsCommand(ListChildBase):
    name = "emobservations"
    description = "List EM observations associated with an event or superevent"


class ListFilesCommand(ListChildBase):
    name = "files"
    description = "List files associated with an event or superevent"


class ListLabelsCommand(ListChildBase):
    name = "labels"
    description = "List labels attached to an event or superevent"


class ListLogsCommand(ListChildBase):
    name = "logs"
    description = "List log entries for an event or superevent"


class ListSignoffsCommand(ListChildBase):
    name = "signoffs"
    description = "List signoffs associated with a superevent"
    long_description = textwrap.dedent("""\
        List signoffs associated with a superevent. Event signoff retrieval
        is not presently implemented.
    """).rstrip()
    parent_parsers = (superevent_id_parser,)


class ListTagsCommand(ListChildBase):
    name = "tags"
    description = "List tags attached to an event or superevent log entry"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "N",
            type=int,
            help="Index number of the log entry"
        )
        return parser

    def run(self, client, args):
        return client.tags(args.object_id, args.N)


class ListVoeventsCommand(ListChildBase):
    name = "voevents"
    description = "List VOEvents associated with an event or superevent"



# -*- coding: utf-8 -*-
# Copyright (C) Brian Moe, Branson Stephens (2015)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
from base64 import b64encode
from cryptography import x509
from cryptography.hazmat.backends import default_backend
import datetime
from io import open
import json
import os
import socket
import ssl
import sys

import six
from six.moves import http_client
from six.moves.urllib.parse import urlparse
import mimetypes


from .exceptions import HTTPError
from .extern.safe_netrc import netrc as safe_netrc
from .version import __version__

DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api/"

# ----------------------------------------------------------------
# HTTP/S Proxy classes
# Taken from: http://code.activestate.com/recipes/456195/


class ProxyHTTPConnection(http_client.HTTPConnection):

    _ports = {'http': 80, 'https': 443}

    def request(self, method, url, body=None, headers={}):
        # request is called before connect, so can interpret url and get
        # real host/port to be used to make CONNECT request to proxy
        o = urlparse(url)
        proto = o.scheme
        port = o.port
        host = o.hostname
        if proto is None:
            raise ValueError("unknown URL type: %s" % url)
        if port is None:
            try:
                port = self._ports[proto]
            except KeyError:
                raise ValueError("unknown protocol for: %s" % url)
        self._real_host = host
        self._real_port = port
        http_client.HTTPConnection.request(self, method, url, body, headers)

    def connect(self):
        http_client.HTTPConnection.connect(self)
        # send proxy CONNECT request
        self.send("CONNECT {0}:{1} HTTP/1.0\r\n\r\n".format(
                  self._real_host, self._real_port))
        # expect a HTTP/1.0 200 Connection established
        response = self.response_class(self.sock, strict=self.strict,
                                       method=self._method)
        (version, code, message) = response._read_status()
        # probably here we can handle auth requests...
        if code != 200:
            # proxy returned and error, abort connection, and raise exception
            self.close()
            raise socket.error("Proxy connection failed: {0} {1}".format(
                               code, message.strip()))
        # eat up header block from proxy....
        while True:
            # should not use directly fp probably
            line = response.fp.readline()
            if line == '\r\n':
                break


class ProxyHTTPSConnection(ProxyHTTPConnection):
    default_port = 443

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 strict=None, context=None):
        ProxyHTTPConnection.__init__(self, host, port)
        self.key_file = key_file
        self.cert_file = cert_file
        self.context = context

    def connect(self):
        ProxyHTTPConnection.connect(self)
        # make the sock ssl-aware
        self.sock = self.context.wrap_socket(self.sock)


# ----------------------------------------------------------------
# Legacy GSI REST Client
class LegacyGsiRest(object):
    def __init__(self, url=DEFAULT_SERVICE_URL, proxy_host=None,
                 proxy_port=3128, cred=None, username=None, password=None,
                 force_noauth=False, fail_if_noauth=False,
                 reload_certificate=False, reload_buffer=300):
        """
        url (:obj:`str`, optional): URL of server API
        proxy_host (:obj:`str`, optional): proxy host
        proxy_port (:obj:`str`, optional): proxy port
        cred (:obj:`tuple` or :obj:`str, optional): a tuple or list of
            (``/path/to/cert/file``, ``/path/to/key/file) or a single path to
            a combined proxy file (if using an X.509 certificate for
            authentication)
        username (:obj:`str`, optional): username for basic auth
        password (:obj:`str`, optional): password for basic auth
        force_noauth (:obj:`bool`, optional): set to True if you want to skip
            credential lookup and use this client as an unauthenticated user
        fail_if_noauth (:obj:`bool`, optional): set to True if you want the
            constructor to fail if no authentication credentials are provided
            or found
        reload_certificate (:obj:`bool`, optional): if ``True``, your
            certificate will be checked before each request whether it is
            within ``reload_buffer`` seconds of expiration, and if so, it will
            be reloaded. Useful for processes which may live longer than the
            certificate lifetime and have an automated method for certificate
            renewal. The path to the new/renewed certificate **must** be the
            same as for the old certificate.
        reload_buffer (:obj:`int`, optional): buffer (in seconds) for reloading
            a certificate in advance of its expiration. Only used if
            ``reload_certificate`` is ``True``.

        Authentication details:
        You can:
            1. Provide a path to an X.509 certificate and key or a single
               combined proxy file
            2. Provide a username and password
        Or:
            The code will look for a certificate in a default location
                (``/tmp/x509up_u%d``, where ``%d`` is your user ID)
            The code will look for a username and password for the specified
                server in ``$HOME/.netrc``
        """
        # Process service URL
        o = urlparse(url)
        host = o.hostname
        port = o.port or 443

        # Store some of this information
        self._server_host = host
        self._server_port = port
        self._proxy_host = proxy_host
        self._proxy_port = proxy_port
        self._reload_certificate = reload_certificate
        self._reload_buffer = reload_buffer

        # Store information about credentials and authentication type
        self.credentials = {}
        self.auth_type = None

        # Fail if conflicting arguments: (fail if no auth, but force no auth)
        if fail_if_noauth and force_noauth:
            err_msg = ('You have provided conflicting parameters to the '
                       'client constructor: fail_if_noauth=True and '
                       'force_noauth=True.')
            raise ValueError(err_msg)

        # Try to get user-provided credentials, if we aren't forcing
        # no authentication
        if not force_noauth:
            credentials_provided = self._process_credentials(
                cred, username, password)

        # If the user didn't provide credentials in the constructor,
        # we try to look up the credentials
        if not force_noauth and not credentials_provided:
            # Look for X509 certificate and key
            cred = self._find_x509_credentials()
            if cred:
                self.credentials['cert_file'], self.credentials['key_file'] = \
                    cred
                self.auth_type = 'x509'
            else:
                # Look for basic auth credentials in .netrc file
                try:
                    basic_auth_tuple = safe_netrc().authenticators(host)
                except IOError:
                    # IOError = no .netrc file found, pass
                    pass
                else:
                    # If credentials were found for host, set them up!
                    if basic_auth_tuple is not None:
                        self.credentials['username'] = basic_auth_tuple[0]
                        self.credentials['password'] = basic_auth_tuple[2]
                        self.auth_type = 'basic'

        if (fail_if_noauth and not self.credentials):
            raise RuntimeError('No authentication credentials found.')

        # If we are using basic auth, construct auth header
        if (self.auth_type == 'basic'):
            user_and_pass = b64encode('{username}:{password}'.format(
                username=self.credentials['username'],
                password=self.credentials['password']).encode()) \
                .decode('ascii')
            self.authn_header = {
                'Authorization': 'Basic {0}'.format(user_and_pass),
            }

        # If we are using X.509 auth, load the certificate with the
        # cryptography.x509 module
        if (self.auth_type == 'x509'):
            self._load_certificate()

        # Construct version header
        self.version_header = {'User-Agent': 'gracedb-client/{version}'.format(
            version=__version__)}

        # Set up SSL context and connector
        self.set_up_connector(host, port, proxy_host, proxy_port)

    def _load_certificate(self):
        if not self.auth_type == 'x509':
            raise RuntimeError("Can't load certificate for "
                               "non-X.509 authentication.")

        # Open cert file and load it as bytes
        with open(self.credentials['cert_file'], 'rb') as cf:
            cert_data = cf.read()

        # Certificates should be PEM, but just in case, we'll try
        # DER if loading a PEM certificate fails
        try:
            self.certificate = x509.load_pem_x509_certificate(
                cert_data, default_backend()
            )
        except ValueError:
            try:
                self.certificate = x509.load_der_x509_certificate(
                    cert_data, default_backend()
                )
            except ValueError:
                raise RuntimeError('Error importing certificate')

    def _check_certificate_expiration(self, reload_buffer=None):
        if reload_buffer is None:
            reload_buffer = self._reload_buffer
        if (self.auth_type != 'x509'):
            raise RuntimeError("Can't check certificate expiration for "
                               "non-X.509 authentication.")
        if not hasattr(self, 'certificate'):
            self._load_certificate()

        # Compare certificate expiration to current time (UTC)
        # (certs use UTC until 2050, see https://tools.ietf.org/html/rfc5280)
        time_to_expire = \
            (self.certificate.not_valid_after - datetime.datetime.utcnow())
        expired = \
            time_to_expire <= datetime.timedelta(seconds=reload_buffer)
        return expired

    def print_certificate_subject(self):
        if not hasattr(self, 'certificate'):
            raise RuntimeError('No certificate found.')
        print(self.certificate.subject.rfc4514_string())

    def print_certificate_expiration_date(self):
        if not hasattr(self, 'certificate'):
            raise RuntimeError('No certificate found.')
        out_str = '{dt} UTC'.format(dt=self.certificate.not_valid_after)
        print(out_str)

    def set_up_connector(self, host, port, proxy_host, proxy_port):
        # Prepare SSL context
        # Fix for python < 2.7.13
        # https://docs.python.org/2/library/ssl.html#ssl.PROTOCOL_TLS
        # This *needs* to be updated. Eventually 2.7.5 support is going
        # to have to stop.
        if sys.version_info >= (2, 7, 13):
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        else:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)

        if (self.auth_type == 'x509'):
            try:
                ssl_context.load_cert_chain(self.credentials['cert_file'],
                                            self.credentials['key_file'])
            except ssl.SSLError:
                msg = ("\nERROR: Unable to load cert/key pair.\n\nPlease "
                       "run ligo-proxy-init or grid-proxy-init again or "
                       "make sure your robot certificate is readable.\n\n")
                self.output_and_die(msg)
        # Load and verify certificates
        ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.check_hostname = True
        # Find the various CA cert bundles stored on the system
        ssl_context.load_default_certs()

        if proxy_host:
            self.connector = lambda: ProxyHTTPSConnection(
                proxy_host, proxy_port, context=ssl_context)
        else:
            self.connector = lambda: http_client.HTTPSConnection(
                host, port, context=ssl_context)

    def _process_credentials(self, cred, username, password):
        """Process credentials provided in the constructor"""
        credentials_provided = False
        if cred:
            if isinstance(cred, (list, tuple)):
                self.credentials['cert_file'], self.credentials['key_file'] = \
                    cred
            else:
                self.credentials['cert_file'] = cred
                self.credentials['key_file'] = cred
            credentials_provided = True
            self.auth_type = 'x509'
        elif username and password:
            self.credentials['username'] = username
            self.credentials['password'] = password
            credentials_provided = True
            self.auth_type = 'basic'
        elif (username is None) ^ (password is None):
            raise RuntimeError('Must provide both username AND password for '
                               'basic auth.')

        return credentials_provided

    def _find_x509_credentials(self):
        """
        Tries to find a user's X509 certificate and key.  Checks environment
        variables first, then expected location for default proxy.
        """
        proxyFile = os.environ.get('X509_USER_PROXY')
        certFile = os.environ.get('X509_USER_CERT')
        keyFile = os.environ.get('X509_USER_KEY')

        if certFile and keyFile:
            return certFile, keyFile

        if proxyFile:
            return proxyFile, proxyFile

        # Try default proxy
        proxyFile = os.path.join('/tmp', "x509up_u%d" % os.getuid())
        if os.path.exists(proxyFile):
            return proxyFile, proxyFile

        # Try default cert/key
        homeDir = os.environ.get('HOME', None)
        if homeDir:
            certFile = os.path.join(homeDir, '.globus', 'usercert.pem')
            keyFile = os.path.join(homeDir, '.globus', 'userkey.pem')

            if os.path.exists(certFile) and os.path.exists(keyFile):
                return certFile, keyFile

    def show_credentials(self, print_output=True):
        """Prints authentication type and credentials information."""
        output = {'auth_type': self.auth_type}
        output.update(self.credentials)

        if print_output:
            print(output)
        else:
            return output

    def get_user_info(self):
        """Get information from the server about your user account."""
        user_info_link = self.links.get('user-info', None)
        if user_info_link is None:
            raise RuntimeError('Server does not provide a user info endpoint')
        return self.get(user_info_link)

    def getConnection(self):
        return self.connector()

    # When there is a problem with the SSL connection or cert authentication,
    # either conn.request() or conn.getresponse() will throw an exception.
    # The following two wrappers are intended to catch these exceptions and
    # return an intelligible error message to the user.
    # A wrapper for getting the response:
    def get_response(self, conn):
        try:
            return conn.getresponse()
        except ssl.SSLError as e:

            if (self.auth_type == 'x509'):
                # Check for a valid user proxy cert.
                expired = self._check_certificate_expiration(reload_buffer=0)

                if expired:
                    msg = ("\nERROR\n\nYour certificate or proxy has "
                           "expired. Please run ligo-proxy-init or "
                           "grid-proxy-init (as appropriate) to generate "
                           "a fresh one.\n\n")
                else:
                    msg = ("\nERROR\n\nYour certificate appears valid, "
                           "but there was a problem establishing a secure "
                           "connection: {e}").format(e=str(e))
            else:
                msg = ("\nERROR\n\nProblem establishing secure connection: "
                       "{e}\n\n").format(e=str(e))
            self.output_and_die(msg)

    # A wrapper for making the request.
    def make_request(self, conn, *args, **kwargs):
        try:
            conn.request(*args, **kwargs)
        except ssl.SSLError as e:
            msg = "\nERROR \n\n"
            msg += "Problem establishing secure connection: %s \n\n" % str(e)
            self.output_and_die(msg)

    def make_request_and_get_response(self, conn, method, url, body=None,
                                      headers={}):

        # For X.509 based auth: if the user has specified to reload the
        # certificate (upon expiration), check the certificate to see if it
        # has expired
        if (self.auth_type == 'x509' and self._reload_certificate):
            cert_expired = self._check_certificate_expiration()
            if cert_expired:
                self._load_certificate()
                self.set_up_connector(
                    self._server_host, self._server_port, self._proxy_host,
                    self._proxy_port
                )
                conn = self.getConnection()

        # Make request
        self.make_request(conn, method, url, body=body, headers=headers)

        # Get response
        response = self.get_response(conn)

        return response

    def request(self, method, url, body=None, headers=None, priming_url=None):
        # Bug in Python (versions < 2.7.1 (?))
        # http://bugs.python.org/issue11898
        # if the URL is unicode and the body of a request is binary,
        # the POST/PUT action fails because it tries to concatenate
        # the two which fails due to encoding problems.
        # Workaround is to cast all URLs to str.
        # This is probably bad in general,
        # but for our purposes, today, this will do.
        url = url and str(url)
        priming_url = priming_url and str(priming_url)
        headers = headers or {}
        conn = self.getConnection()

        # Add version string to user-agent header
        headers.update(self.version_header)

        # Add auth header for basic auth
        if (self.auth_type == 'basic'):
            headers.update(self.authn_header)

        # Set up priming URL for certain requests using X509 auth
        if (self.auth_type == 'x509' and priming_url):
            priming_header = {'connection': 'keep-alive'}
            priming_header.update(self.version_header)
            response = self.make_request_and_get_response(
                conn, "GET", priming_url, headers=priming_header
            )
            if response.status != 200:
                response = self.adjustResponse(response)
            else:
                # Throw away the response and make sure to read the body.
                response = response.read()

        response = self.make_request_and_get_response(
            conn, method, url, body=body, headers=headers
        )

        # Special handling of 401 unauthorized response for basic auth
        # to catch expired passwords
        if (self.auth_type == 'basic' and response.status == 401):
            try:
                msg = "\nERROR: {e}\n\n".format(json.loads(
                    response.read())['detail'])
            except Exception:
                msg = "\nERROR:\n\n"
            msg += ("\nERROR:\n\nPlease check the username/password in your "
                    ".netrc file. If your password is more than a year old, "
                    "you will need to use the web interface to generate a new "
                    "one.\n\n")
            self.output_and_die(msg)
        return self.adjustResponse(response)

    def adjustResponse(self, response):
        # XXX WRONG.
        if response.status >= 400:
            response_content = response.read()
            if isinstance(response_content, bytes):
                response_content = response_content.decode()
            if response.getheader('x-throttle-wait-seconds', None):
                try:
                    rdict = json.loads(response_content)
                    rdict['retry-after'] = response.getheader(
                        'x-throttle-wait-seconds')
                    response_content = json.dumps(rdict)
                except Exception:
                    pass
            raise HTTPError(response.status, response.reason, response_content)
        response.json = lambda: self.load_json_or_die(response)
        return response

    def get(self, url, headers=None):
        return self.request("GET", url, headers=headers)

    def head(self, url, headers=None):
        return self.request("HEAD", url, headers=headers)

    def delete(self, url, headers=None):
        return self.request("DELETE", url, headers=headers)

    def options(self, url, headers=None):
        return self.request("OPTIONS", url, headers=headers)

    def post(self, *args, **kwargs):
        return self.post_or_put_or_patch("POST", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.post_or_put_or_patch("PUT", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.post_or_put_or_patch("PATCH", *args, **kwargs)

    def post_or_put_or_patch(self, method, url, body=None, headers=None,
                             files=None):
        headers = headers or {}
        if not files:
            # Simple urlencoded body
            if isinstance(body, dict):
                # XXX What about the headers in the params?
                if 'content-type' not in headers:
                    headers['content-type'] = "application/json"
                body = json.dumps(body)
        else:
            body = body or {}
            if isinstance(body, dict):
                body = list(body.items())
            content_type, body = encode_multipart_formdata(body, files)
            # XXX What about the headers in the params?
            headers = {
                'content-type': content_type,
                'content-length': str(len(body)),
                # 'connection': 'keep-alive',
            }
        return self.request(method, url, body, headers)

    # A utility for writing out an error message to the user and then stopping
    # execution. This seems to behave sensibly in both the interpreter and in
    # a script.
    @classmethod
    def output_and_die(cls, msg):
        raise RuntimeError(msg)

    # Given an HTTPResponse object, try to read its content and interpret as
    # JSON--or die trying.
    @classmethod
    def load_json_or_die(cls, response):

        # First check that the response object actually exists.
        if not response:
            raise ValueError("No response object")

        # Next, try to read the content of the response.
        response_content = response.read()
        response_content = response_content.decode('utf-8')
        if not response_content:
            response_content = '{}'

        # Finally, try to create a dict by decoding the response as JSON.
        rdict = None
        try:
            rdict = json.loads(response_content)
        except ValueError:
            msg = "ERROR: got unexpected content from the server:\n"
            msg += response_content
            raise ValueError(msg)

        return rdict


# ----------------------------------------------------------------
# HTTP upload encoding
# Taken from http://code.activestate.com/recipes/146306/
def encode_multipart_formdata(fields, files):
    """
    fields is a sequence of (name, value) elements for regular form fields.
    files is a sequence of (name, filename, value) elements for data to be
        uploaded as files.
    Returns (content_type, body) ready for httplib.HTTP instance
    """
    BOUNDARY = '----------ThIs_Is_tHe_bouNdaRY_$'
    CRLF = b'\r\n'
    L = []
    for (key, value) in fields:
        if value is None:
            continue
        L.append(('--' + BOUNDARY).encode())
        L.append('Content-Disposition: form-data; name="{0}"'.format(
            key).encode())
        L.append(''.encode())
        # encode in case it is unicode
        if isinstance(value, six.binary_type):
            L.append(value)
        elif isinstance(value, six.text_type):
            L.append(value.encode())
        else:
            L.append(str(value).encode())
    for (key, filename, value) in files:
        if value is None:
            continue
        L.append(('--' + BOUNDARY).encode())
        # str(filename) in case it is unicode
        L.append(('Content-Disposition: form-data; name="{0}"; '
                 'filename="{1}"').format(key, filename).encode())
        L.append('Content-Type: {0}'.format(get_content_type(filename))
                 .encode())
        L.append(''.encode())
        if isinstance(value, six.text_type):
            L.append(value.encode())
        else:
            L.append(value)
    L.append(('--' + BOUNDARY + '--').encode())
    L.append(''.encode())
    body = CRLF.join(L)
    content_type = 'multipart/form-data; boundary={0}'.format(BOUNDARY) \
        .encode()
    return content_type, body


def get_content_type(filename):
    return mimetypes.guess_type(filename)[0] or 'application/octet-stream'



# Tests for intermediate commands
#  Ex: 'gracedb add' because it has subcommands like 'gracedb add event'
#  NOT 'gracedb confirm_as_gw' because it has no subcommands.
import re
import pytest
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Test data ###################################################################
###############################################################################
COMMANDS_TEST_DATA = [
    ("add", ["event", "label", "tag"]),
    ("create", ["emobservation", "event", "log", "signoff", "superevent",
                "voevent"]),
    ("delete", ["signoff"]),
    ("get", ["emobservation", "event", "file", "label", "log", "signoff",
             "superevent", "voevent"]),
    ("list", ["emobservations", "files", "labels", "logs", "signoffs",
              "tags", "voevents"]),
    ("remove", ["event", "label", "tag"]),
    ("search", ["events", "superevents"]),
    ("update", ["event", "signoff", "superevent", "grbevent"]),
]


###############################################################################
# Tests #######################################################################
###############################################################################
@pytest.mark.parametrize("cmd, expected", COMMANDS_TEST_DATA)
def test_subcommands(CLI, cmd, expected):
    """
    Ensure that the subcommands for the specified level 1 command are
    as expected
    """
    cmd_instance = CLI.subcommand_dict[cmd]()

    assert sorted(list(cmd_instance.subcommand_dict)) == sorted(expected)


@pytest.mark.parametrize("cmd", [sc[0] for sc in COMMANDS_TEST_DATA])
def test_bad_subcommand(CLI, capsys, cmd):
    """Test handling of bad subcommand input to level 1 commands"""
    fake_sc = 'fake_subcommand'
    with pytest.raises(SystemExit) as excinfo:
        CLI([cmd, fake_sc])

    # Check exit code
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out
    assert "{cmd} {sc}' not found".format(cmd=cmd, sc=fake_sc) in stdout


@pytest.mark.parametrize("cmd", [sc[0] for sc in COMMANDS_TEST_DATA])
def test_command_help(CLI, cmd):
    """Test level 1 command help functionality"""
    print_help_func_str = \
        'ligo.gracedb.cli.parsers.CustomHelpArgumentParser.print_help'
    with mock.patch(print_help_func_str) as mock_print_help:
        with pytest.raises(SystemExit) as excinfo:
            CLI([cmd, '--help'])

    # Make sure print_help is called once
    assert mock_print_help.call_count == 1

    # Check exit code
    assert excinfo.value.code == 0


@pytest.mark.parametrize("cmd,expected", COMMANDS_TEST_DATA)
def test_command_help_content(CLI, capsys, cmd, expected):
    """Check level 1 command help content"""
    with pytest.raises(SystemExit) as excinfo:
        CLI([cmd, '--help'])

    # Check exit code
    assert excinfo.value.code == 0

    # Get content of help message
    stdout = capsys.readouterr().out

    # Verify entry for each command
    cmd_entry_regex = r'\s+{cmd}\s+'
    available_commands = stdout[stdout.index('Available commands'):]
    for cmd in expected:
        assert re.search(cmd_entry_regex.format(cmd=cmd), available_commands)



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize(
    "method",
    ['signoffs', 'create_signoff', 'update_signoff', 'delete_signoff']
)
def test_event_signoff_actions(safe_client, method):
    graceid = 'T123456'
    err_msg = 'Not yet implemented for events'
    with pytest.raises(NotImplementedError, match=err_msg):
        getattr(safe_client, method)(
            graceid, 'fake_arg1', 'fake_arg2', 'fake_arg3'
        )


@pytest.mark.parametrize(
    "signoff_type,instrument",
    [(None, ""), ("OP", "H1"), ("ADV", "")]
)
def test_get_superevent_signoffs(safe_client, signoff_type, instrument):
    superevent_id = 'TS120345zd'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if signoff_type:
        template_key = 'superevent-signoff-detail-template'
    else:
        template_key = 'superevent-signoff-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb._signoff_helper') as mock_sh, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.signoffs(superevent_id, signoff_type=signoff_type,
                             instrument=instrument)

    call_args, call_kwargs = mock_sh.call_args
    assert len(call_args) == 4
    assert call_args == (superevent_id, 'get', mock_template,
                         {'superevent_id': superevent_id},)
    assert len(call_kwargs) == 2
    assert call_kwargs['signoff_type'] == signoff_type
    assert call_kwargs['instrument'] == instrument


@pytest.mark.parametrize(
    "signoff_type,instrument",
    [(None, ""), ("OP", "H1"), ("ADV", "")]
)
def test_create_superevent_signoff(safe_client, signoff_type, instrument):
    superevent_id = 'TS120345zd'
    status = 'OK'
    comment = 'test comment'

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-signoff-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb._signoff_helper') as mock_sh, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.create_signoff(superevent_id, signoff_type, status,
                                   comment, instrument=instrument)

    call_args, call_kwargs = mock_sh.call_args
    assert len(call_args) == 4
    assert call_args == (superevent_id, 'create', mock_template,
                         {'superevent_id': superevent_id},)
    assert len(call_kwargs) == 4
    assert call_kwargs['signoff_type'] == signoff_type
    assert call_kwargs['instrument'] == instrument
    assert call_kwargs['status'] == status
    assert call_kwargs['comment'] == comment


@pytest.mark.parametrize(
    "signoff_type,instrument",
    [(None, ""), ("OP", "H1"), ("ADV", "")]
)
def test_update_superevent_signoff(safe_client, signoff_type, instrument):
    superevent_id = 'TS120345zd'
    status = 'OK'
    comment = 'test comment'

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-signoff-detail-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb._signoff_helper') as mock_sh, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.update_signoff(superevent_id, signoff_type, status,
                                   comment, instrument=instrument)

    call_args, call_kwargs = mock_sh.call_args
    assert len(call_args) == 4
    assert call_args == (superevent_id, 'update', mock_template,
                         {'superevent_id': superevent_id},)
    assert len(call_kwargs) == 4
    assert call_kwargs['signoff_type'] == signoff_type
    assert call_kwargs['instrument'] == instrument
    assert call_kwargs['status'] == status
    assert call_kwargs['comment'] == comment


@pytest.mark.parametrize("status,comment", [(None, None), (None, "")])
def test_update_signoff_bad_args(safe_client, status, comment):
    err_msg = "Provide at least one of 'status' or 'comment'"
    with pytest.raises(ValueError, match=err_msg):
        safe_client.update_signoff('TS120412bc', 'ADV', status=status,
                                   comment=comment)


@pytest.mark.parametrize(
    "signoff_type,instrument",
    [(None, ""), ("OP", "H1"), ("ADV", "")]
)
def test_delete_superevent_signoff(safe_client, signoff_type, instrument):
    superevent_id = 'TS120345zd'

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-signoff-detail-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb._signoff_helper') as mock_sh, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.delete_signoff(superevent_id, signoff_type,
                                   instrument=instrument)

    call_args, call_kwargs = mock_sh.call_args
    assert len(call_args) == 4
    assert call_args == (superevent_id, 'delete', mock_template,
                         {'superevent_id': superevent_id},)
    assert len(call_kwargs) == 2
    assert call_kwargs['signoff_type'] == signoff_type
    assert call_kwargs['instrument'] == instrument


@pytest.mark.parametrize(
    "action,signoff_type,instrument,status,comment",
    [
        ('create', 'ADV', '', 'NO', 'test comment'),
        ('update', 'OP', 'H1', None, None),
        ('update', 'OP', 'H1', 'OK', None),
        ('update', 'OP', 'H1', None, 'new comment'),
        ('update', 'ADV', '', 'OK', 'looks good'),
        ('get', None, None, None, None),
        ('get', 'OP', 'L1', None, None),
        ('get', 'ADV', '', None, None),
        ('delete', 'OP', 'V1', None, None),
        ('delete', 'ADV', '', None, None),
    ]
)
def test_signoff_helper(safe_client, action, signoff_type, instrument, status,
                        comment):
    superevent_id = 'TS340229zzz'

    # Mocks
    template = mock.MagicMock()
    uri_kwargs = {}

    # Set up mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {
        'signoff-types': {'ADV': 'advocate', 'OP': 'operator'},
        'instruments': {'H1': 'Hanford', 'L1': 'Livingston', 'V1': 'Virgo'},
        'signoff-statuses': {'OK': 'OKAY', 'NO': 'NOT OKAY'},
    }

    # Mock request method
    expected_request_method_dict = {
        'create': 'POST',
        'update': 'PATCH',
        'get': 'GET',
        'delete': 'DELETE',
    }
    request_func = 'ligo.gracedb.rest.GraceDb.{action}'.format(
        action=expected_request_method_dict[action].lower())

    # Call
    with mock.patch(request_func) as mock_rf, \
         mock.patch(si_property, mock_si_dict):  # noqa: E127
        safe_client._signoff_helper(
            superevent_id, action, template, uri_kwargs,
            signoff_type=signoff_type, instrument=instrument, status=status,
            comment=comment
        )

    # Check template kwargs
    template_args, template_kwargs = template.format.call_args
    assert len(template_args) == 0
    if action == 'create':
        assert len(template_kwargs) == 0
    else:
        if signoff_type:
            assert len(template_kwargs) == 1
            assert template_kwargs['typeinst'] == signoff_type + instrument
        else:
            assert len(template_kwargs) == 0

    # Check request method call
    rf_args, rf_kwargs = mock_rf.call_args
    assert len(rf_args) == 1
    if action == 'create':
        assert len(rf_kwargs) == 1
        assert 'data' in rf_kwargs
        assert rf_kwargs['data']['signoff_type'] == signoff_type
        assert rf_kwargs['data']['instrument'] == instrument
        assert rf_kwargs['data']['status'] == status
        assert rf_kwargs['data']['comment'] == comment
    elif action == 'update' and (status or comment):
        assert len(rf_kwargs) == 1
        assert 'data' in rf_kwargs
        body_len = 0
        if comment:
            body_len += 1
            assert rf_kwargs['data']['comment'] == comment
        if status:
            body_len += 1
            assert rf_kwargs['data']['status'] == status
        assert len(rf_kwargs['data']) == body_len
    else:
        assert len(rf_kwargs) == 0


def test_signoff_helper_bad_signoff_type(safe_client):

    # Set up mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {
        'signoff-types': {'ADV': 'advocate', 'OP': 'operator'},
    }

    err_msg = 'signoff_type must be one of: {st}'.format(
        st=", ".join(mock_si_dict['signoff-types']))
    with mock.patch(si_property, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client._signoff_helper('TS340229zz', '', '', '',
                                        signoff_type='FAKE')


def test_signoff_helper_bad_instrument(safe_client):

    # Set up mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {
        'instruments': {'H1': 'Hanford', 'L1': 'Livingston', 'V1': 'Virgo'},
    }

    err_msg = 'instrument must be one of: {inst}'.format(
        inst=", ".join(mock_si_dict['instruments']))
    with mock.patch(si_property, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client._signoff_helper('TS340229zz', '', '', '',
                                        instrument='FAKE')


def test_signoff_helper_bad_status(safe_client):

    # Set up mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {
        'signoff-statuses': {'OK': 'OKAY', 'NO': 'NOT OKAY'},
    }

    err_msg = 'status must be one of: {status}'.format(
        status=", ".join(mock_si_dict['signoff-statuses']))
    with mock.patch(si_property, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client._signoff_helper('TS340229zz', '', '', '',
                                        status='FAKE')


def test_signoff_helper_operator_signoff_no_instrument(safe_client):

    # Set up mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {
        'signoff-types': {'ADV': 'advocate', 'OP': 'operator'},
    }

    err_msg = 'Operator signoffs require an instrument'
    with mock.patch(si_property, mock_si_dict):
        with pytest.raises(ValueError, match=err_msg):
            safe_client._signoff_helper('TS340229zz', '', '', '',
                                        signoff_type='OP')


def test_signoff_helper_bad_action(safe_client):

    err_msg = "action should be 'create', 'update', 'get', or 'delete'"
    with pytest.raises(ValueError, match=err_msg):
        safe_client._signoff_helper('TS340229zz', 'fake', '', '')



# Tests for subcommands below 'add'
#  Ex: 'gracedb add event'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_add_event_subcommand(CLI):
    """Test add event subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'g_id': 'G1234',
    }
    cmd = 'add event {s_id} {g_id}'.format(**cmd_args)

    func = 'ligo.gracedb.rest.GraceDb.addEventToSuperevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['g_id'],)
    assert cli_kwargs == {}


def test_add_label_subcommand(CLI):
    """Test add label subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'label': 'FAKE_LABEL',
    }
    cmd = 'add label {s_id} {label}'.format(**cmd_args)

    func = 'ligo.gracedb.rest.GraceDb.writeLabel'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['label'],)
    assert cli_kwargs == {}


TAG_DISPLAY_NAMES = [None, 'test_display_name']
@pytest.mark.parametrize("display_name", TAG_DISPLAY_NAMES)  # noqa: E302
def test_add_tag_subcommand(CLI, display_name):
    """Test add tag subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'log_N': 12,
        'tag_name': 'FAKE_TAG',
    }
    cmd = 'add tag {s_id} {log_N} {tag_name}'.format(**cmd_args)
    if display_name is not None:
        cmd += ' --tag-display-name={dn}'.format(dn=display_name)

    func = 'ligo.gracedb.rest.GraceDb.addTag'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['log_N'],)
    assert cli_kwargs == {'tag_name': cmd_args['tag_name'],
                          'displayName': display_name}



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock
import os

import pytest


@pytest.mark.parametrize("filename", ['file.txt', "", None])
def test_superevent_files(safe_client, filename):
    superevent_id = 'S190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if filename:
        template_key = 'superevent-file-detail-template'
    else:
        template_key = 'superevent-file-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.files(superevent_id, filename=filename)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    if filename:
        stream_args = {'stream': True}
    else:
        stream_args = {}
    assert get_call_kwargs == stream_args

    # Test template call kwargs
    num_kwargs = 1
    if filename:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['superevent_id'] == superevent_id
    if filename:
        assert template_call_kwargs['file_name'] == filename


@pytest.mark.parametrize("filename", ['file.txt', "", None])
def test_event_files(safe_client, filename):
    graceid = 'T123456'

    # Set up templates mock
    mock_template = mock.MagicMock()
    template_key = 'files-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.files(graceid, filename=filename)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    if filename:
        stream_args = {'stream': True}
    else:
        stream_args = {}
    assert get_call_kwargs == stream_args

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 2
    assert template_call_kwargs['graceid'] == graceid
    if not filename:
        filename = ""
    assert template_call_kwargs['filename'] == filename


@pytest.mark.parametrize(
    "filename,filecontents",
    [
        ('file.txt', None),
        ('subdir1/subdir2/file.txt', None),
        ('-', None),
        ('file.txt', 'fc1'),
        ('subdir1/subdir2/file.txt', 'fc2'),
        ('-', 'fc3'),
    ]
)
def test_write_log_with_file(safe_client, filename, filecontents):
    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-log-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up mock open
    open_func = 'ligo.gracedb.rest.open'
    mock_data = 'fake data'
    open_mocker = mock.mock_open(read_data=mock_data)

    # Set up mock sys.stdin.read
    stdin_obj = 'ligo.gracedb.rest.sys.stdin'
    mock_stdin_data = 'fake stdin data'

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(open_func, open_mocker), \
         mock.patch(stdin_obj) as mock_stdin, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127

        mock_stdin.read.return_value = mock_stdin_data
        safe_client.writeLog('TS121212a', 'test', filename=filename,
                             filecontents=filecontents)

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 2
    assert 'data' in post_call_kwargs
    assert 'files' in post_call_kwargs

    # Test file contents
    post_files = post_call_kwargs['files']
    assert len(post_files) == 1
    assert 'upload' in post_files
    post_files = post_files['upload']
    if filecontents is None:
        if filename == '-':
            assert post_files[0] == 'stdin'
            assert post_files[1] == mock_stdin_data
        else:
            assert post_files[0] == os.path.basename(filename)
            assert post_files[1] == open_mocker()
    else:
        assert post_files[0] == os.path.basename(filename)
        assert post_files[1] == filecontents


def test_write_log_with_filecontents_handler(safe_client):
    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-log-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up mock filecontents
    filename = 'file.txt'
    filecontents = mock.MagicMock()
    mock_data = 'more fake data'
    filecontents.read.return_value = mock_data

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127

        safe_client.writeLog('TS121212a', 'test', filename=filename,
                             filecontents=filecontents)

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 2
    assert 'data' in post_call_kwargs
    assert 'files' in post_call_kwargs

    # Test file contents
    post_files = post_call_kwargs['files']
    assert len(post_files) == 1
    assert 'upload' in post_files
    post_files = post_files['upload']
    assert post_files[0] == os.path.basename(filename)
    assert post_files[1] == filecontents
    assert post_files[2] == 'text/plain'



import pytest
import sys
try:
    from unittest import mock
except ImportError:  # py < 3
    import mock

from ligo.gracedb.rest import GraceDb


def test_provide_x509_cert_and_key():
    """Test client instantiation with provided certificate and key files"""
    # Set up cert and key files
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    load_cert_func = 'ligo.gracedb.cert.load_certificate'
    with mock.patch(load_cert_func):  # noqa: E127
        # Initialize client
        g = GraceDb(cred=(cert_file, key_file))

    # Output credentials:
    creds = g.show_credentials(print_output=False)

    # Check credentials
    assert len(g.cert) == 2
    assert g.auth_type == 'x509'
    assert creds.get('cert_file') == cert_file
    assert creds.get('key_file') == key_file


def test_provide_x509_proxy():
    """Test client instantiation with provided combined proxy file"""
    # Set up combined proxy file
    proxy_file = '/tmp/proxy_file'

    exists_func = 'os.path.exists'
    load_cert_func = 'ligo.gracedb.cert.load_certificate'
    with mock.patch(load_cert_func), \
         mock.patch(exists_func) as mock_exists_func:  # noqa: E127

        mock_exists_func.return_value = True
        # Initialize client
        g = GraceDb(cred=proxy_file)

    # Output credentials:
    creds = g.show_credentials(print_output=False)

    # Check credentials
    assert type(g.cert) == str
    assert g.auth_type == 'x509'
    assert creds.get('cert_file') == proxy_file
    assert creds.get('key_file') == proxy_file


USER_PASS_TEST_DATA = [
    {'username': 'user', 'password': 'pw'},
    {'username': 'user'},
    {'password': 'pwd'},
]
@pytest.mark.parametrize("user_and_or_pass", USER_PASS_TEST_DATA)  # noqa: E302
def test_provide_username_and_password(user_and_or_pass):
    """Test client instantiation with provided username and or password"""

    # Initialize client - should fail if only one of username and
    # password is provided
    if not len(user_and_or_pass) == 2:
        err_str = 'You must provide both a username and a password '\
                  'for basic authentication'
        with pytest.raises(RuntimeError, match=err_str):
            g = GraceDb(**user_and_or_pass)
    else:
        g = GraceDb(**user_and_or_pass)

        # Output credentials:
        creds = g.show_credentials(print_output=False)

        assert len(g.auth) == 2
        assert g.auth_type == 'basic'
        assert creds.get('username') == \
            user_and_or_pass.get('username')
        assert creds.get('password') == \
            user_and_or_pass.get('password')


def test_provide_all_creds():
    """Test providing all credentials to the constructor"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'
    username = 'user'
    password = 'pw'

    # Initialize client
    g = GraceDb(
        cred=(cert_file, key_file), username=username, password=password
    )

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert len(g.cert) == 2
    assert g.auth_type == 'x509'
    assert creds.get('cert_file') == cert_file
    assert creds.get('key_file') == key_file


def test_x509_credentials_lookup():
    """Test lookup of X509 credentials"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client
    find_x509_func = 'ligo.gracedb.client.GraceDBClient._get_x509_credentials'
    with mock.patch(find_x509_func) as mock_find_x509:  # noqa: E127
        mock_find_x509.return_value = (cert_file, key_file)
        g = GraceDb()

    # the 'auth_type' variable doesn't get set in this case, but it's
    # sort of fictitious when self.cert is being mocked.
    assert len(g.cert) == 2


def test_x509_lookup_cert_key_from_envvars():
    """Test lookup of X509 cert and key from environment variables"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client
    fake_creds_dict = {
        'X509_USER_CERT': cert_file,
        'X509_USER_KEY': key_file,
    }
    os_environ_func = 'os.environ'
    with mock.patch.dict(os_environ_func, fake_creds_dict):  # noqa: E127
        g = GraceDb()

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert len(g.cert) == 2
    assert g.auth_type == 'x509'
    assert creds.get('cert_file') == cert_file
    assert creds.get('key_file') == key_file


def test_x509_lookup_proxy_from_envvars():
    """Test lookup of X509 combined provxy file from environment variables"""
    # Setup
    proxy_file = '/tmp/proxy_file'

    # Initialize client
    os_environ_func = 'os.environ'
    mock_environ_dict = {'X509_USER_PROXY': proxy_file}
    with mock.patch.dict(os_environ_func, mock_environ_dict):  # noqa: E127
        g = GraceDb()

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert type(g.cert) == str
    assert g.auth_type == 'x509'
    assert creds.get('cert_file') == proxy_file
    assert creds.get('key_file') == proxy_file


def test_basic_credentials_lookup():
    """Test client instantiation - look up basic auth creds from .netrc file"""
    # Set up credentials and mock_netrc return
    fake_creds = {
        'machine': 'fake.com',
        'login': 'fake_user',
        'password': 'fake_password',
    }
    netrc = 'ligo.gracedb.client.netrc'
    find_x509_func = 'ligo.gracedb.client.GraceDBClient._get_x509_credentials'
    with mock.patch(find_x509_func) as mock_find_x509, \
         mock.patch(netrc) as mock_netrc:  # noqa: E127

        # Force lookup to not find any X509 credentials
        mock_find_x509.return_value = None
        # Mock return value from netrc lookup
        mock_netrc().authenticators.return_value = (
            fake_creds['login'], None, fake_creds['password'])

        # Initialize client
        g = GraceDb('https://{0}/api/'.format(fake_creds['machine']))

        # Check credentials
        creds = g.show_credentials(print_output=False)

        assert len(g.auth) == 2
        assert g.auth_type == 'basic'
        assert creds.get('username') == fake_creds.get('login')
        assert creds.get('password') == fake_creds.get('password')


@pytest.mark.parametrize("fail_if_noauth", [True, False])
def test_no_credentials(fail_if_noauth):
    """Test client instantiation with no credentials at all"""
    find_x509_func = 'ligo.gracedb.client.GraceDBClient._get_x509_credentials'
    netrc = 'ligo.gracedb.client.netrc'
    with mock.patch(find_x509_func) as mock_find_x509, \
         mock.patch(netrc) as mock_netrc:  # noqa: E127

        # Force lookup to not find any X509 credentials
        mock_find_x509.return_value = None
        # Mock return value from netrc lookup
        mock_netrc().authenticators.return_value = None

        # Initialize client
        if fail_if_noauth:
            err_str = 'No authentication credentials could be found'
            with pytest.raises(RuntimeError, match=err_str):
                g = GraceDb(fail_if_noauth=fail_if_noauth)
        else:
            if sys.version_info[0] > 2:
                warn_str = 'Authentication credentials not found, proceeding'
                with pytest.warns(UserWarning, match=warn_str):
                    g = GraceDb(fail_if_noauth=fail_if_noauth)
            else:
                g = GraceDb(fail_if_noauth=fail_if_noauth)

            # Check credentials
            assert g.auth_type is None


def test_force_noauth():
    """Test forcing no authentication, even with X509 certs available"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client
    environ_dict = 'os.environ'
    mock_environ_dict = {
        'X509_USER_CERT': cert_file,
        'X509_USER_KEY': key_file,
    }
    with mock.patch.dict(environ_dict, mock_environ_dict):  # noqa: E127

        # Initialize client
        g = GraceDb(force_noauth=True)

    # Check credentials
    assert g.auth_type is None


@pytest.mark.parametrize("creds_found", [True, False])
def test_fail_if_noauth(creds_found):
    """Test failing if no authentication credentials are provided"""
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client
    find_x509_func = 'ligo.gracedb.client.GraceDBClient._get_x509_credentials'
    netrc = 'ligo.gracedb.client.netrc'
    with mock.patch(find_x509_func) as mock_find_x509, \
         mock.patch(netrc) as mock_netrc:  # noqa: E127

        if not creds_found:
            # Force lookup to not find any X509 credentials
            mock_find_x509.return_value = None
            # Mock return value from netrc lookup
            mock_netrc().authenticators.return_value = None

            # Initialize client
            err_str = 'No authentication credentials could be found'
            with pytest.raises(RuntimeError, match=err_str):
                g = GraceDb(fail_if_noauth=True)
        else:
            # Initialize client:
            mock_find_x509.return_value = (cert_file, key_file)
            g = GraceDb(cred=(cert_file, key_file), fail_if_noauth=True)

            # Check credentials
            assert len(g.cert) == 2


def test_force_noauth_and_fail_if_noauth():
    # Initialize client
    err_str = ('You have provided conflicting parameters to the client '
               'constructor: fail_if_noauth=True and force_noauth=True.')
    with pytest.raises(ValueError, match=err_str):
        GraceDb(force_noauth=True, fail_if_noauth=True)


@pytest.mark.parametrize(
    "resource,key",
    [
        ('api_versions', 'api-versions'),
        ('server_version', 'server-version'),
        ('links', 'links'),
        ('templates', 'templates'),
        ('groups', 'groups'),
        ('pipelines', 'pipelines'),
        ('searches', 'searches'),
        ('allowed_labels', 'labels'),
        ('superevent_categories', 'superevent-categories'),
        ('em_groups', 'em-groups'),
        ('voevent_types', 'voevent-types'),
        ('signoff_types', 'signoff-types'),
        ('signoff_statuses', 'signoff-statuses'),
        ('instruments', 'instruments'),
    ]
)
def test_properties_from_api_root(safe_client, resource, key):
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    with mock.patch(si_prop, new_callable=mock.PropertyMock()) as mock_si:
        getattr(safe_client, resource)

    call_args, call_kwargs = mock_si.get.call_args
    assert mock_si.get.call_count == 1
    assert len(call_args) == 1
    assert call_kwargs == {}
    assert call_args[0] == key


@pytest.mark.parametrize("api_version", [1, 1.2, [], (), {}])
def test_bad_api_version(api_version):
    err_msg = 'api_version should be a string'
    with pytest.raises(TypeError, match=err_msg):
        GraceDb(api_version=api_version)


@pytest.mark.parametrize(
    "service_url,api_version",
    [
        ('test', None),
        ('test/', None),
        ('test', 'v1'),
        ('test/', 'v2'),
        ('test/', 'default'),
    ],
)
def test_set_service_url(safe_client, service_url, api_version):
    safe_client._set_service_url(service_url, api_version)

    # Construct expected service urls
    expected_service_url = service_url.rstrip('/') + '/'
    expected_versioned_service_url = expected_service_url

    if api_version and api_version != 'default':
        expected_versioned_service_url += api_version + '/'

    assert safe_client._service_url == expected_service_url
    assert safe_client._versioned_service_url == expected_versioned_service_url



import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import object_id_parser, superevent_id_parser, graceid_parser


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class RemoveCommand(RegisteredCommandBase):
    name = "remove"
    description = textwrap.dedent("""\
        Remove a label from an event or superevent, a tag from a log
        entry, or an event from a superevent
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class RemoveChildBase(RegisteredSubCommandBase):
    _registry = registry


class RemoveLabelCommand(RemoveChildBase):
    name = "label"
    description = "Remove a label from an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument("label", type=str, help="Name of label to remove")
        return parser

    def run(self, client, args):
        return client.removeLabel(args.object_id, args.label)


class RemoveTagCommand(RemoveChildBase):
    name = "tag"
    description = "Remove a tag from a log entry"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "log_number",
            type=int,
            help="Index number of log entry"
        )
        parser.add_argument("tag_name", type=str, help="Name of tag to remove")
        return parser

    def run(self, client, args):
        return client.removeTag(args.object_id, args.log_number, args.tag_name)


class RemoveEventCommand(RemoveChildBase):
    name = "event"
    description = "Remove an event from a superevent"
    parent_parsers = (superevent_id_parser, graceid_parser,)

    def run(self, client, args):
        return client.removeEventFromSuperevent(args.superevent_id,
                                                args.graceid)



# Instructions for adding new commands
# Most are double-level commands (i.e., gracedb create event [params]) vs.
# single-level commands (gracedb expose [params])

# The command registry is populated by a custom metaclass when new commands
# are created. The commands must be imported here to ensure that they are added
# to the registry.
# Note that this command_registry is imported in ligo.gracedb.cli.commands.base
# so that it can be populated there.
command_registry = []


from .add import AddCommand
from .create import CreateCommand
from .delete import DeleteCommand
from .get import GetCommand
from .list import ListCommand
from .remove import RemoveCommand
from .search import SearchCommand
from .update import UpdateCommand

# Other single-level commands
from .subcommands import (
    ConfirmAsGwCommand, CredentialsCommand, ExposeCommand, HideCommand,
    PingCommand, InfoCommand, ShowCommand,
)


__all__ = ['command_registry']



import os

import pytest


# Apply module-level mark
pytestmark = pytest.mark.integration


def test_create_gstlal(client, test_data_dir):
    # Create gstlal event
    filename = os.path.join(test_data_dir, 'cbc-lm.xml')
    response = client.createEvent('Test', 'gstlal', filename)
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'gstlal'
    assert data['gpstime'] == 971609248.151741
    assert data['extra_attributes']['CoincInspiral']['snr'] == 9.31793628458239


def test_create_gstlal_with_missing_table_entry(client, test_data_dir):
    # Create gstlal event, with a missing snglinspiral table
    # entry. The event should get made and the table populated with
    # a None entry.
    filename = os.path.join(test_data_dir, 'cbc-lm-missing-entry.xml')
    response = client.createEvent('Test', 'gstlal', filename)
    assert response.status_code == 201
    data = response.json()

    # Make sure warnings are empty:
    assert data['warnings'] == []

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'gstlal'
    assert data['gpstime'] == 971609248.151741
    assert data['extra_attributes']['CoincInspiral']['snr'] == 9.31793628458239

    # Extra results. Assert that the snr snglinspiral table entry is empty
    assert data['extra_attributes']['SingleInspiral'][0].get('snr') is None


def test_create_gstlal_no_ilwdchar(client, test_data_dir):
    # Create gstlal event, where the ligol ilwd:char entries
    # have been replaced with int8's.
    filename = os.path.join(test_data_dir, 'cbc-lm-no-ilwdchar.xml')
    response = client.createEvent('Test', 'gstlal', filename)
    assert response.status_code == 201
    data = response.json()

    # Make sure warnings are empty:
    assert data['warnings'] == []

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'gstlal'
    assert data['gpstime'] == 971609248.151741
    # Also, this assert will fail if the event is not read correctly:
    assert data['extra_attributes']['CoincInspiral']['snr'] == 9.31793628458239


@pytest.mark.xfail
def test_create_pycbc(client, test_data_dir):
    # Create PyCBC event - need to get a PyCBC data file and clean it up,
    # but the files are very large, so this would be quite an undertaking
    raise NotImplementedError()


def test_create_mbta(client, test_data_dir):
    # Create MBTA event
    filename = os.path.join(test_data_dir, 'cbc-mbta.xml')
    response = client.createEvent('Test', 'MBTAOnline', filename)
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'MBTAOnline'
    assert data['gpstime'] == 1078903329.421037
    assert data['extra_attributes']['CoincInspiral']['mass'] == \
        2.963776448988085


def test_create_spiir(client, test_data_dir):
    # Create SPIIR event
    filename = os.path.join(test_data_dir, 'spiir-test.xml')
    response = client.createEvent('Test', 'spiir', filename)
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert response.status_code == 201
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'spiir'
    assert data['extra_attributes']['CoincInspiral']['mass'] == 3.98
    assert data['far'] == 3.27e-07


def test_create_spiir_no_ilwdchar(client, test_data_dir):
    # Create SPIIR event, with ilwd:char entires converted to int8s
    filename = os.path.join(test_data_dir, 'spiir-test-no-ilwdchar.xml')
    response = client.createEvent('Test', 'spiir', filename)
    assert response.status_code == 201
    data = response.json()

    # Make sure warnings are empty:
    assert data['warnings'] == []

    # Test results
    assert response.status_code == 201
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'spiir'
    assert data['extra_attributes']['CoincInspiral']['mass'] == 3.98
    assert data['far'] == 3.27e-07


def test_create_cwb(client, test_data_dir):
    # Create CWB event
    filename = os.path.join(test_data_dir, 'burst-cwb.txt')
    response = client.createEvent('Test', 'CWB', filename)
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'CWB'
    assert data['far'] == 0.00019265
    assert data['extra_attributes']['MultiBurst']['amplitude'] == 5.017162


def test_create_olib(client, test_data_dir):
    # Create oLIB event
    filename = os.path.join(test_data_dir, 'olib-test.json')
    response = client.createEvent('Test', 'oLIB', filename)
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'oLIB'
    assert data['far'] == 7.22e-06
    assert data['extra_attributes']['LalInferenceBurst']['bci'] == 1.111
    assert data['extra_attributes']['LalInferenceBurst']['bsn'] == 7.19


def test_create_hardwareinjection(client, test_data_dir):
    # Create hardware injection event
    inst = 'H1'
    filename = os.path.join(test_data_dir, 'sim-inj.xml')
    response = client.createEvent(
        'Test', 'HardwareInjection', filename, instrument=inst,
        source_channel="", destination_channel=""
    )
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'HardwareInjection'
    assert data['instruments'] == inst


def test_create_external(client, test_data_dir):
    # Create external GRB event
    filename = os.path.join(test_data_dir, 'fermi-test.xml')
    response = client.createEvent('Test', 'Fermi', filename, search='GRB')
    assert response.status_code == 201
    data = response.json()

    # Test results
    assert data['group'] == 'Test'
    assert data['pipeline'] == 'Fermi'
    assert data['extra_attributes']['GRB']['error_radius'] == 8.8374
    assert data['extra_attributes']['GRB']['ra'] == 345.99


@pytest.mark.parametrize("offline", [True, False])
def test_creation_with_offline(client, test_data_dir, offline):
    filename = os.path.join(test_data_dir, 'cbc-lm.xml')
    response = client.createEvent('Test', 'gstlal', filename, offline=offline)
    assert response.status_code == 201
    data = response.json()
    assert data['offline'] == offline


def test_creation_with_labels(client, test_data_dir):
    labels = ['INJ', 'DQV']
    filename = os.path.join(test_data_dir, 'cbc-lm.xml')
    response = client.createEvent('Test', 'gstlal', filename, labels=labels)
    assert response.status_code == 201
    data = response.json()
    assert len(data['labels']) == len(labels)
    for l in labels:
        assert l in data['labels']


def test_replace_event(client, test_data_dir):
    # Create an event
    filename = os.path.join(test_data_dir, 'cbc-lm.xml')
    response = client.createEvent('Test', 'gstlal', filename)
    assert response.status_code == 201
    data1 = response.json()
    graceid = data1['graceid']

    # Check gpstime
    assert data1['gpstime'] == 971609248.151741

    # Replace it
    new_data_file = os.path.join(test_data_dir, 'cbc-lm2.xml')
    response = client.replaceEvent(graceid, new_data_file)
    assert response.status_code == 202

    # Get event data and check that gpstime has changed
    response = client.event(graceid)
    assert response.status_code == 200
    data2 = response.json()
    assert data2['gpstime'] == 971609249.151741


def test_get_event(client, test_data_dir):
    # Create gstlal event
    filename = os.path.join(test_data_dir, 'cbc-lm.xml')
    response = client.createEvent('Test', 'gstlal', filename)
    assert response.status_code == 201
    data1 = response.json()
    graceid = data1['graceid']

    # Get event and compare
    response = client.event(graceid)
    assert response.status_code == 200
    data2 = response.json()
    for k in data2:
        assert data1[k] == data2[k]


def test_search(client, create_event):
    # Create an event
    event = create_event()

    # Search by ID
    response = client.events(event['graceid'])
    results_list = list(response)
    assert len(results_list) == 1
    assert results_list[0]['graceid'] == event['graceid']



import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import superevent_id_parser, graceid_parser


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class UpdateCommand(RegisteredCommandBase):
    name = "update"
    description = textwrap.dedent("""\
        Update an event with a new event file, update a superevent's
        parameters, update a GRB event's parameters, or update an existing
        superevent signoff
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class UpdateChildBase(RegisteredSubCommandBase):
    _registry = registry


class UpdateEventCommand(UpdateChildBase):
    name = "event"
    description = "Update an event by uploading a new event data file"
    long_description = textwrap.dedent("""\
        Replace the initial data file which was uploaded to create an event.
        This will cause the event's parameters to be updated based on the newly
        provided data file.
    """).rstrip()
    parent_parsers = (graceid_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "filename",
            type=str,
            help="Path to new event file"
        )
        return parser

    def run(self, client, args):
        return client.replaceEvent(args.graceid, args.filename)


class UpdateSignoffCommand(UpdateChildBase):
    name = "signoff"
    description = "Update a superevent signoff"
    long_description = textwrap.dedent("""\
        Update a superevent signoff with a new status and/or comment.
        Event signoff is not presently implemented.
    """).rstrip()

    parent_parsers = (superevent_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "signoff_type",
            type=str,
            help=("Signoff type (do '{prog} info signoff_types') to see "
                  "options").format(prog=self.base_prog)
        )
        parser.add_argument(
            "instrument",
            type=str,
            nargs='?',
            help=("Instrument code (do {prog} info instruments to see "
                  "options). Required for operator signoffs.")
            .format(prog=self.base_prog)
        )
        parser.add_argument(
            "--status",
            type=str,
            help=("Signoff status (do '{prog} info signoff_statuses') to see "
                  "options").format(prog=self.base_prog)
        )
        parser.add_argument(
            "--comment",
            type=str,
            help="Comment to update the signoff with"
        )
        return parser

    def run(self, client, args):
        instrument = args.instrument or ''  # Convert None to ''
        return client.update_signoff(
            args.superevent_id, args.signoff_type, status=args.status,
            comment=args.comment, instrument=instrument
        )


class UpdateSupereventCommand(UpdateChildBase):
    name = "superevent"
    description = "Update a superevent's parameters"
    long_description = textwrap.dedent("""\
        Update t_start, t_0, t_end, and/or the preferred_event for a superevent
    """).rstrip()
    parent_parsers = (superevent_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            '--t-start',
            type=float,
            help="New t_start value for superevent"
        )
        parser.add_argument(
            '--t-0',
            type=float,
            help="New t_0 value for superevent"
        )
        parser.add_argument(
            '--t-end',
            type=float,
            help="New t_end value for superevent"
        )
        parser.add_argument(
            '--preferred-event',
            type=str,
            help="Graceid of new preferred event for superevent"
        )
        return parser

    def run(self, client, args):
        return client.updateSuperevent(
            args.superevent_id, t_0=args.t_0, t_start=args.t_start,
            t_end=args.t_end, preferred_event=args.preferred_event
        )


class UpdateGrbEventCommand(UpdateChildBase):
    name = "grbevent"
    description = "Update GRB event-specific parameters"
    long_description = textwrap.dedent("""\
        Update ra, dec, error_radius, t90, redshift, and/or declination for
        a GRB event.
    """).rstrip()
    parent_parsers = (graceid_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            '--ra',
            type=float,
            help="New right ascension value for the GRB event (degrees)"
        )
        parser.add_argument(
            '--dec',
            type=float,
            help="New declination value for the GRB event (degrees)"
        )
        parser.add_argument(
            '--error-radius',
            type=float,
            help="New error radius value for the GRB event (degrees)"
        )
        parser.add_argument(
            '--t90',
            type=float,
            help="New t90 value for the GRB event (seconds)"
        )
        parser.add_argument(
            '--redshift',
            type=float,
            help="New redshift value for the GRB event"
        )
        parser.add_argument(
            '--designation',
            type=str,
            help=("New designation for the GRB event (GRByymmddx or "
                  "GRByymmddfff format)")
        )
        return parser

    def run(self, client, args):
        return client.update_grbevent(
            args.graceid, ra=args.ra, dec=args.dec,
            error_radius=args.error_radius, t90=args.t90,
            redshift=args.redshift, designation=args.designation
        )



import json
import pytest
# import six
try:
    from unittest import mock
except ImportError:
    import mock

from ligo.gracedb.exceptions import HTTPError
from requests import Response

# Apply module-level mark
pytestmark = pytest.mark.cli


@pytest.mark.parametrize("output_type", ['json', 'status'])
def test_main_with_ok_response(main_tester, capsys, output_type):
    """Test CLI wrapper when response object is returned with good status"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:

        # Set up response from CLI.__call__
        status = 200
        reason = 'because'
        response_mock = mock.MagicMock(
            spec=Response,
            status_code=status, reason=reason)
        mock_CLI().return_value = response_mock
        # Set up response.json()
        json_mock = mock.MagicMock()
        json_mock.return_value = {'test': 'ok'}
        # Use setattr since the MagicMock is autospecced
        setattr(response_mock, 'json', json_mock)

        # Set up CLI.args.output_type
        mock_CLI().args.output_type = output_type

        # Call function
        main_tester()

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    if output_type == 'json':
        assert json.loads(stdout) == json_mock.return_value
    elif output_type == 'status':
        assert stdout.rstrip() == 'Server returned {status}: {reason}'.format(
            status=status, reason=reason)


MSG_DATA = ['short message', 'long' * 500]
@pytest.mark.parametrize("response_msg", MSG_DATA,  # noqa: E302
                         ids=['short', 'long'])
def test_main_with_bad_response(main_tester, capsys, response_msg):
    """Test CLI wrapper when response object is returned with status >= 400"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:

        # Set up response from CLI.__call__
        status = 400
        reason = 'Bad Request'
        response_mock = mock.MagicMock(
            spec=Response,
            status_code=status, reason=reason,
            text=response_msg)
        mock_CLI().return_value = response_mock

        # Set up CLI.args.output_type
        mock_CLI().args.output_type = 'json'

        # Call function
        with pytest.raises(SystemExit) as excinfo:
            main_tester()

    # Make sure exit code is correct
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    if len(response_msg) < 1000:
        assert stdout.rstrip() == '{code} {reason}: {msg}'.format(
            code=status, reason=reason, msg=response_msg)
    else:
        assert stdout.rstrip() == '{code} {reason}'.format(
            code=status, reason=reason)


# Define an error response object:
err_response = Response()
err_response.status_code = 400
err_response.reason = 'Bad Request'
err_response._content = b'test'
EXCEPTION_DATA = [Exception('test'),
                  HTTPError(response=err_response)]


@pytest.mark.parametrize("exc", EXCEPTION_DATA)  # noqa: E302
def test_main_with_exception_in_CLI_call(main_tester, capsys, exc):
    """Test CLI wrapper when CLI call raises exception"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:
        # Set up response from CLI.__call__
        mock_CLI().side_effect = exc

        # Call function
        with pytest.raises(SystemExit) as excinfo:
            main_tester()

    # Make sure exit code is correct
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    if isinstance(exc, HTTPError):
        assert stdout.rstrip() == 'Error: {code} {reason}. {text}.'.format(
            code=exc.status_code, reason=exc.reason, text=exc.text)
    elif isinstance(exc, Exception):
        assert stdout.rstrip() == 'Error: {msg}'.format(msg=str(exc))


def test_main_with_exception_in_response_json(main_tester, capsys):
    """Test CLI wrapper when response.json() raises exception"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:
        # Set up response from CLI.__call__
        status = 200
        reason = 'OK'
        response_mock = mock.MagicMock(
            spec=Response,
            status_code=status, reason=reason)
        mock_CLI().return_value = response_mock
        # Set up response.read()
        exc_msg = 'uh oh'
        json_mock = mock.MagicMock()
        json_mock.side_effect = Exception(exc_msg)
        # Use setattr since the MagicMock is autospecced
        setattr(response_mock, 'json', json_mock)

        # Set up CLI.args.output_type
        mock_CLI().args.output_type = 'json'

        # Call function
        with pytest.raises(SystemExit) as excinfo:
            main_tester()

    # Make sure exit code is correct
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    assert stdout.rstrip() == exc_msg


def test_main_with_str_response(main_tester, capsys):
    """Test CLI wrapper with string response"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:
        # Set up response from CLI.__call__
        str_response = 'this is a string response'
        mock_CLI().return_value = str_response

        # Call function
        main_tester()

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    assert stdout.rstrip() == str_response


def test_main_with_unexpected_response_type(main_tester, capsys):
    """Test CLI wrapper with unexpected response type"""

    CLI = 'ligo.gracedb.cli.client.CommandLineInterface'
    with mock.patch(CLI) as mock_CLI:
        # Set up response from CLI.__call__
        response = float(1.2345)
        mock_CLI().return_value = response

        # Call function
        with pytest.raises(SystemExit) as excinfo:
            main_tester()

    # Make sure exit code is correct
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out

    # Test output
    expected = "Unexpected response type {tp}\nResponse: {resp}".format(
        tp=type(response), resp=str(response))
    assert stdout.rstrip() == expected



# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Tanner Prestegard,
#               Branson Stephens, Brian Moe (2020)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>

# This file contains x509 certificate loading and checking tools.
# Mostly duplicated effort from Tanner's work in support of certificate
# reloading, but it could be useful for giving the client the option
# of checking their certificates, or displaying certificate validity and
# expiration dates when they use the 'gracedb credentials client' command.

from cryptography import x509
from cryptography.hazmat.backends import default_backend
import datetime

far_future = datetime.timedelta(days=365)


# Takes in the path of a certificate and retrrns an x509.cryptography
# certificate object. This removes the prior check that the auth_type
# be 'x509' since this should only be called when reload_certificate
# is true, which checks for x509, or in cases in general when users want
# to check a cert without connecting a client.

def load_certificate(cert):
    """ Loads in a path to a x509 certificate and returns a
        x509.cryptography object """
    with open(cert, 'rb') as cert_obj:
        cert_obj = cert_obj.read()

    # Try loading with PEM, then try loading with DER, then give up.

    try:
        return x509.load_pem_x509_certificate(
            cert_obj, default_backend()
        )
    except ValueError:
        try:
            return x509.load_pem_x509_certificate(
                cert_obj, default_backend()
            )
        except ValueError:
            raise RuntimeError('Error importing certificate')


# Checks certificate expiration and returns a boolean. Optionally checks
# that the time left until expiration within the 'reload_buffer' parameter.

def check_certificate_expiration(cert, reload_buffer=0):
    """ Checks to see if a cert is expiring within an optional
        reload_buffer parameter. Default, reload_buffer=0, meaning
        is the certificate currently expired.

        cert is either a string path to an x509 cert, or a x509
        cryptography certificate object """

    if not hasattr(cert, 'subject'):
        cert = load_certificate(cert)

    expired = (cert.not_valid_after - datetime.datetime.utcnow()) <= \
        datetime.timedelta(seconds=reload_buffer)

    return expired
